;; Generated package description from flylisp.el  -*- no-byte-compile: t -*-
(define-package "flylisp" "0.2.0.20220923.134141" "Color unbalanced parentheses and parentheses inconsistent with indentation" '((emacs "24.1") (cl-lib "0.4")) :commit "da6e87169f9c28bc4fce88adfefb6d49a240d692" :url "https://elpa.gnu.org/packages/flylisp.html" :authors '(("Barry O'Reilly" . "gundaetiapo@gmail.com")) :maintainer '("Barry O'Reilly" . "gundaetiapo@gmail.com"))
